import datetime
from os.path import exists, basename, join, dirname

from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QAction
from osgeo import gdal
from qgis._core import QgsProject, QgsRasterLayer
from qgis._gui import QgisInterface


def classFactory(iface):
    return Plugin(iface)


class Plugin(object):

    def __init__(self, iface):
        assert isinstance(iface, QgisInterface)
        self.iface = iface
        self.icon = QIcon(join(dirname(__file__), 'icon.png'))

    def initGui(self):
        self.action = QAction(self.icon, 'One Click Raster Stacking', self.iface.mainWindow())
        self.action.triggered.connect(self.run)
        self.iface.addToolBarIcon(self.action)

    def unload(self):
        self.iface.removeToolBarIcon(self.action)

    def run(self):

        selectedLayers = self.iface.layerTreeView().selectedLayers()

        if len(selectedLayers) != 1:
            self.iface.messageBar().pushCritical(
                'Quick Raster Stacking usage error', 'Select a single raster layer defining the CRS for stacking.'
            )
            return

        selectedLayer = selectedLayers[0]
        if not isinstance(selectedLayer, QgsRasterLayer):
            self.iface.messageBar().pushCritical(
                'Quick Raster Stacking usage error', 'Select a single raster layer defining the CRS for stacking.'
            )

        timestamp = str(datetime.datetime.now()).replace(' ', 'T').replace(':', '')
        filenames = list()
        bandnames = list()
        i1 = 0
        for layer in QgsProject.instance().mapLayers().values():
            if not isinstance(layer, QgsRasterLayer):
                continue
            if not exists(layer.source()):
                continue
            if layer.crs() != selectedLayer.crs():
                continue
            i1 += 1

            ds: gdal.Dataset = gdal.Open(layer.source())
            for i2 in range(ds.RasterCount):
                rb: gdal.Band = ds.GetRasterBand(i2 + 1)
                filename = f"/vsimem/raster_{i1}_band_{i2 + 1}_{timestamp}.vrt"
                options = gdal.TranslateOptions(bandList=[i2 + 1], format='VRT')
                gdal.Translate(filename, ds, options=options)
                filenames.append(filename)
                bandname = rb.GetDescription()
                if bandname == '':
                    bandname = f'Band {i2 + 1}'
                bandnames.append(f'{layer.name()} - {bandname}')

        if len(filenames) == 0:
            self.iface.messageBar().pushCritical(
                'Quick Raster Stacking', 'No matching raster found.'
            )
            return

        filename = f"/vsimem/stack_{timestamp}.vrt"
        ds = gdal.BuildVRT(filename, filenames, options=gdal.BuildVRTOptions(separate=True, resolution='highest'))
        for i2, bandname in enumerate(bandnames):
            rb: gdal.Band = ds.GetRasterBand(i2 + 1)
            rb.SetDescription(bandname)
        ds.FlushCache()
        del ds

        self.iface.addRasterLayer(filename, basename(filename))
